--[[
    This file is part of darktable,
    This is a rewritten script. The original script geotools.lua and clear_GPS.lua was written by Tobias Jakobs
    copyright (c) 2016 Tobias Jakobs
    copyright (C) 2016 Bill Ferguson
    copyright (c) 2017 Holger Klemm
    
    darktable is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    darktable is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with darktable.  If not, see <http://www.gnu.org/licenses/>.
]]
--[[

USAGE
* require this script from your main lua file
  To do this add this line to the file .config/darktable/luarc: 
require "geodata_editor"

* it creates a new geodata_editor lighttable module
]]

local dt = require "darktable"
local df = require "lib/dtutils.file"
local gettext = dt.gettext

dt.configuration.check_version(...,{3,0,0},{4,0,0})

-- Tell gettext where to find the .mo file translating messages for a particular domain
gettext.bindtextdomain("geodata_editor",dt.configuration.config_dir.."/lua/")

local function _(msgid)
    return gettext.dgettext("geodata_editor", msgid)
end

-- GUI
button_copy = dt.new_widget("button")
{
 label = _('copy geodata'),
 tooltip = _('copy geodata'),
 clicked_callback = function (_)  
   copy_gps()
 end
}


local check_button_latitude = dt.new_widget("check_button")
{
  label = _('latitude [decimal degrees]'), 
  value = true,
  reset_callback = function(self) 
       self.value = true
    end
}

local entry_latitude = dt.new_widget("entry")
{
  text = "", 
  placeholder = "",
  editable = true,
  tooltip = _('laitude position'),
  reset_callback = function(self) 
       self.text = "" 
    end
}

local check_button_longitude = dt.new_widget("check_button")
{
  label = _('longitude [decimal degrees]'), 
  value = true,
  reset_callback = function(self) 
       self.value = true
    end
}

local entry_longitude = dt.new_widget("entry")
{
  text = "", 
  placeholder = "",
  editable = true,
  tooltip = _('longitude position'),
  reset_callback = function(self) 
       self.text = "" 
    end
}


local check_button_elevation = dt.new_widget("check_button")
{
  label = _('elevation [meter]'), 
  value = true,
  reset_callback = function(self) 
       self.value = true
    end
}

local entry_elevation = dt.new_widget("entry")
{
  text = "", 
  placeholder = "",
  editable = true,
  tooltip = _('elevation'),
  reset_callback = function(self) 
       self.text = "" 
    end
}

button_paste = dt.new_widget("button")
{
  label = _('paste geodata'),
  tooltip = _('paste geodata'),
  clicked_callback = function (_)  
     paste_gps(images)
  end
}

button_remove = dt.new_widget("button")
{
  label = _('remove geodata'),
  tooltip = _('remove geodata'),
  clicked_callback = function (_)  
     clear_GPS(dt.gui.action_images)
  end
}

local separator1 = dt.new_widget("separator")
{
}
--[[
local button_start_app = dt.new_widget("button")
{
  label = _("start geo app"),
  tooltip = _("start map or google earth"),
  clicked_callback = open_location_in_gnome_maps
}
]]--
local label_calc_dist = dt.new_widget("label")
{
  label = _('distance between two selected geo images:'),
  ellipsize = "start",
  halign = "start"
}

local label_distance = dt.new_widget("label")
{
  label = _('distance'),
  ellipsize = "start",
  halign = "start"
}

local separator2 = dt.new_widget("separator")
{
}


local label_altitude = dt.new_widget("label")
{
  label = _('export altitude specification [CSV]'),
  ellipsize = "start",
  halign = "start"
}

local label_filename = dt.new_widget("label")
{
  label = _('CSV filename'),
  ellipsize = "start",
  halign = "start"
}

local altitude_filename = dt.new_widget("entry")
{
  text = "altitude.csv", 
  placeholder = "altitude.csv",
  editable = true,
  tooltip = _('name of the exported file'),
  reset_callback = function(self) 
    self.text = "altitude.csv" 
  end
}

local altitude_file_chooser_button = dt.new_widget("file_chooser_button")
{
  title = _('export altitude specification [CSV]'),  -- The title of the window when choosing a file
  value = "",                     -- The currently selected file
  is_directory = true             -- True if the file chooser button only allows directories to be selecte
}


local button_export_altitude = dt.new_widget("button")
{
  label = _('export altitude specification'),
  tooltip = _('create an altitude profile using the GPS data in the metadata'),
  clicked_callback = altitude_profile
}







-- FUNCTIONS        
        
local function select_with_gps()
  local selection = {}
  job = dt.gui.create_job(_('selecting images with geodata'), true, stop_selection)

  for key,image in ipairs(dt.collection) do
    if job.valid then
      job.percent = (key-1)/#dt.collection

      if (image.longitude and image.latitude) then
        table.insert(selection,image)
      end
    else
      break
    end
  end

  dt.gui.selection(selection)
  job.valid = false
end

local function select_without_gps()
  local selection = {}
  job = dt.gui.create_job(_('selecting images without geodata'), true, stop_selection)

  for key,image in ipairs(dt.collection) do
    if job.valid then
      job.percent = (key-1)/#dt.collection

      if (not image.longitude and not image.latitude) then
        table.insert(selection,image)
      end
    else
      break
    end
  end
  dt.gui.selection(selection)
  job.valid = false
end

local function stop_selection(job)
    job.valid = false
end

-- This is used for older images in the DB
function isnan(x) return x ~= x end

-- Function from:
-- https://forums.coronalabs.com/topic/29019-convert-string-to-date/
-- This looks much scarier than it really is. The goal is to turn that string 
-- into a Unix timestamp (number of seconds since Jan 1, 1970, the standard 
-- used by most systems). So we use the string.match() method to fetch the 
-- various date and time parts into their own variables: xyear, xmonth,etc.

        
 --[[       
        
function make_time_stamp(dateString)
  local convertedTimestamp
--dt.print_error(dateString)
  if (dateString) then
    local pattern = "(%d+)%:(%d+)%:(%d+) (%d+):(%d+):(%d+)"
    local xyear, xmonth, xday, xhour, xminute, xseconds = dateString:match(pattern)
--dt.print_error(xyear)
    convertedTimestamp = os.time({
        year = xyear, 
        month = xmonth, 
        day = xday, 
        hour = xhour, 
        min = xminute, 
        sec = xseconds})
  else
    convertedTimestamp = 0
  end

  return convertedTimestamp
end

local first_have_data = false
local first_latitude = ''
local first_longitude = ''
local first_elevation = ''
local first_image_date = 0

local function get_first_coordinate()
  local sel_images = dt.gui.action_images

  first_latitude = ''    
  first_longitude = ''
  first_elevation = ''
  first_image_date = 0

  for _,image in ipairs(sel_images) do
    if not image then
      first_have_data = false
    else
      image_date = image.exif_datetime_taken
      first_have_data = true
      if (image.latitude) then
        first_latitude = image.latitude    
      end
      if (image.longitude) then
          first_longitude = image.longitude
      end
      if (image.elevation) then
          first_elevation = image.elevation
      end
      if (image.exif_datetime_taken) then
          first_image_date = make_time_stamp(image.exif_datetime_taken)
--dt.print_error(image.exif_datetime_taken)
--dt.print_error(first_image_date)      
      end
    end
    return
  end
end

local second_have_data = false
local second_latitude = ''
local second_longitude = ''
local second_elevation = ''
local second_image_date = 0

local function get_second_coordinate()
  local sel_images = dt.gui.action_images

  second_latitude = ''    
  second_longitude = ''
  second_elevation = ''
  second_image_date = 0

  for _,image in ipairs(sel_images) do
    if not image then
      second_have_data = false
    else
      image_date = image.exif_datetime_taken
      second_have_data = true
      if (image.latitude) then
        second_latitude = image.latitude    
      end
      if (image.longitude) then
          second_longitude = image.longitude
      end
      if (image.elevation) then
          second_elevation = image.elevation
      end
      if (image.exif_datetime_taken) then
          second_image_date = make_time_stamp(image.exif_datetime_taken)
      end
    end
    return
  end
end

local calc_in_between_slider = dt.new_widget("slider")
{
  label = "Position between", 
  soft_min = 0,      -- The soft minimum value for the slider, the slider can't go beyond this point
  soft_max = 100,     -- The soft maximum value for the slider, the slider can't go beyond this point
  hard_min = -100,       -- The hard minimum value for the slider, the user can't manually enter a value beyond this point
  hard_max = 200,    -- The hard maximum value for the slider, the user can't manually enter a value beyond this point
  value = 50          -- The current value of the slider
}

--ToDo: this needs more love
local function calc_in_between()
  local sel_images = dt.gui.action_images
  for _,image in ipairs(sel_images) do
    if image then
      image_date = make_time_stamp(image.exif_datetime_taken)
      if (first_have_data and second_have_data) then
        local start_new      = 0
        local end_new        = second_image_date - first_image_date
        local image_date_new = image_date - first_image_date

        local percent_in_between
        if (end_new == 0) then
          percent_in_between = 1
        else
          percent_in_between = image_date_new/end_new
        end        
        calc_in_between_slider.value = percent_in_between * 100
dt.print_error(percent_in_between)
        local in_between_latitude  = first_latitude + (second_latitude - first_latitude) * percent_in_between
        local in_between_longitude = first_longitude + (second_longitude - first_longitude) * percent_in_between

        if (first_elevation and second_elevation) then
          local in_between_elevation = first_elevation + (second_elevation - first_elevation) * percent_in_between

dt.print_error(first_image_date)
dt.print_error(second_image_date)
dt.print_error(image_date)
dt.print_error(first_elevation)
dt.print_error(second_elevation)
dt.print_error(in_between_elevation)

        end
      end
    end
  end
end
]]--


local copy_gps_have_data = false
local copy_gps_latitude = ''
local copy_gps_longitude = ''
local copy_gps_elevation = ''

function copy_gps()
    button_copy.sensitive=false
  local sel_images = dt.gui.action_images

  copy_gps_latitude = ''    
  copy_gps_longitude = ''
  copy_gps_elevation = ''

  for _,image in ipairs(sel_images) do
    if not image then
      copy_gps_have_data = false
    else
      copy_gps_have_data = true
      if (image.latitude and check_button_latitude.value) then
        copy_gps_latitude = image.latitude 
        entry_latitude.text = image.latitude 
      end
      if (image.longitude and check_button_longitude.value) then
          copy_gps_longitude = image.longitude
          entry_longitude.text = image.longitude
      end
      if (image.elevation and check_button_elevation.value) then
          copy_gps_elevation = image.elevation
          entry_elevation.text = image.elevation
      end
    end
      button_copy.sensitive = true
    return
  end
   button_copy.sensitive = true
end


function paste_gps(images)
    button_paste.sensitive=false
    
  local sel_images = dt.gui.action_images
local NaN = 0/0
  for _,image in ipairs(sel_images) do
    if (check_button_latitude.value) then
        if (entry_latitude.text == "") then
        image.latitude = NaN 
        else
        image.latitude = string.gsub(entry_latitude.text ,"%.",",")
        end
    end
    if (check_button_longitude.value) then   
        if (entry_longitude.text == "") then
        image.longitude = NaN
        else
        image.longitude = string.gsub(entry_longitude.text ,"%.",",")
        end
    end
    if (check_button_elevation.value) then
        if (entry_elevation.text == "") then
        image.elevation = NaN
        else
        image.elevation = string.gsub(entry_elevation.text ,"%.",",")
        end
    end
  end
  dt.print(_('paste finished'))
  button_paste.sensitive=true
end


-- not a number
local NaN = 0/0
function clear_GPS(images)   
    button_remove.sensitive=false
  for _, image in ipairs(images) do
    -- set the location information to Not a Number (NaN) so it displays correctly
    if (check_button_latitude.value) then
     image.latitude = NaN
    end    
    if (check_button_longitude.value) then   
      image.longitude = NaN  
    end    
     if (check_button_elevation.value) then   
    image.elevation = NaN
     end
  end
  dt.print(_('geo data has been removed from the database, but not in the xmp files (darktable bug)'))
      button_remove.sensitive=true
end

--[[
local function open_location_in_gnome_maps()

  if not df.check_if_bin_exists("gnome-maps") then
    dt.print_error(_("gnome-maps not found"))
    return
  end	
    
  local sel_images = dt.gui.action_images
  
  local lat1 = 0;
  local lon1 = 0;
  local i = 0;
  
  -- Use the first image with geo information
  for _,image in ipairs(sel_images) do
    if ((image.longitude and image.latitude) and 
        (image.longitude ~= 0 and image.latitude ~= 90) -- Sometimes the north-pole but most likely just wrong data
       ) then
        lat1 = image.latitude;
        lon1 = image.longitude;
        break
    end

    local startCommand
    startCommand = "gnome-maps \"geo:" .. lat1 .. "," .. lon1 .."\""
    dt.print_error(startCommand)
    
    if coroutine.yield("RUN_COMMAND", startCommand) then
      dt.print(_("Command failed ..."))
    end

  end

end
]]--

-- I used code from here:
-- http://stackoverflow.com/questions/27928/how-do-i-calculate-distance-between-two-latitude-longitude-points
local function get_distance(lat1, lon1, ele1, lat2, lon2, ele2)

    local earthRadius = 6371; -- Radius of the earth in km
    local dLat = math.rad(lat2-lat1);  -- deg2rad below
    local dLon = math.rad(lon2-lon1); 
    local a = 
      math.sin(dLat/2) * math.sin(dLat/2) +
      math.cos(math.rad(lat1)) * math.cos(math.rad(lat2)) * 
      math.sin(dLon/2) * math.sin(dLon/2)
      ;
    local angle = 2 * math.atan2(math.sqrt(a), math.sqrt(1-a)); 
    local distance = earthRadius * angle; -- Distance in km  

    -- Add the elevation to the calculation
    local elevation = 0
    elevation = math.abs(ele1 - ele2) / 1000;  --in km

    if (elevation > 0) then
      distance = math.sqrt(math.pow(elevation,2) + math.pow(distance,2) )
    end
    return distance
end

local function calc_distance()

	  local lat1 = 0;
    local lon1 = 0;
    local lat2 = 0;
    local lon2 = 0;
    local ele1 = 0;
    local ele2 = 0;
    local i = 0;

    local sel_images = dt.gui.selection()

    for _,image in ipairs(sel_images) do
      if ((image.longitude and image.latitude) and
            (image.longitude ~= 0 and image.latitude ~= 90) -- Sometimes the north-pole but most likely just wrong data
           ) then
          
        i = i + 1;
          
        if (i == 1) then
          lat1 = image.latitude
          lon1 = image.longitude
          if (image.elevation) then
            ele1 = image.elevation
          else
            ele1 = 0
          end
        end
         
        lat2 = image.latitude
        lon2 = image.longitude
        if (image.elevation) then
          ele2 = image.elevation
        else
          ele2 = 0;
        end

      end
    end

    local distance = get_distance(lat1, lon1, ele1, lat2, lon2, ele2)

    if (distance < 1) then
        distance = distance * 1000
        distanceUnit = _("m")
    else
        distanceUnit = _("km")
    end
    
    return string.format(_("distance %.2f %s"), distance, distanceUnit)
end

local function print_calc_distance()
    dt.print(calc_distance())
end

local function toolbox_calc_distance()
    label_distance.label = calc_distance()
end



local function altitude_profile()
	  dt.print(_('start export'))
    local sel_images = dt.gui.action_images

    local lat1 = 0;
    local lon1 = 0;
    local lat2 = 0;
    local lon2 = 0;
    local ele1 = 0;
    local ele2 = 0;
    local i = 0;
    local csv_file = '';
    csv_file = "km;m".."\n";

    local distance = 0;
    local distanceFromStart = 0;
    local elevation = 0;
    local elevationAdd = 0;

    local sel_images = dt.gui.action_images
    for _,image in ipairs(sel_images) do
      if ((not isnan(image.longitude) and not isnan(image.latitude) and not isnan(image.elevation) and image.elevation) and 
            (image.longitude ~= 0 and image.latitude ~= 90) -- Sometimes the north-pole but most likely just wrong data
           ) then
          
        i = i + 1;
          
        if (i == 1) then
          lat1 = image.latitude
          lon1 = image.longitude
          ele1 = image.elevation
        else
          lat1 = lat2
          lon1 = lon2
          ele1 = ele2
        end
         
        lat2 = image.latitude
        lon2 = image.longitude
        ele2 = image.elevation

        if (i == 1) then
          distanceFromStart = 0
        else
          local distance = get_distance(lat1, lon1, ele1, lat2, lon2, ele2)
          distanceFromStart = distanceFromStart + distance;
        end

        csv_file = csv_file .. distanceFromStart..";"..image.elevation.."\n";

      end

    end --for

    local exportDirectory = altitude_file_chooser_button.value
    local exportFilename = altitude_filename.text
    if (exportFilename == '') then
      exportFilename = altitude_filename.placeholder
    end
    file = io.open(exportDirectory.."/"..exportFilename, "w")
    file:write(csv_file)
    file:close()
    dt.print(_('file created in ')..exportDirectory)

end




-- REGISTER

dt.register_lib(
  "geodata_editor",        -- Module name
  _('geodata editor'),       -- name
  true,                -- expandable
  true,               -- resetable
  {[dt.gui.views.lighttable] = {"DT_UI_CONTAINER_PANEL_RIGHT_CENTER", 100}},   -- containers
  dt.new_widget("box")
  {
    orientation = "vertical",
    check_button_latitude,
    entry_latitude,
    check_button_longitude,
    entry_longitude,
    check_button_elevation,
    entry_elevation,
    button_copy,
    button_paste,
    button_remove,
    separator1,--------------------------------------------------------
 --   button_start_app,
    label_calc_dist,
    label_distance,
    separator2,--------------------------------------------------------
    label_altitude,
    label_filename,
    altitude_filename,
    altitude_file_chooser_button,
    button_export_altitude,
  },
  nil,-- view_enter
  nil -- view_leave
)




dt.gui.libs.select.register_selection(_('select all with geodata'),
  function(event, images) select_with_gps() end,
  "select all with geodata")

dt.gui.libs.select.register_selection(_('select all without geodata'),
  function(event, images) select_without_gps() end,
  "select all without geodata")

  
dt.gui.libs.image.register_action(
  _('copy geodata'),
  function(event, images) copy_gps() end,
  "copy geodata")  
  
dt.gui.libs.image.register_action(
  _('paste geodata'),
  function(event, images) paste_gps() end,
  "paste geodata")    

  
dt.gui.libs.image.register_action(
  _('remove geodata'),
  function(event, images) clear_GPS(images) end,
  "remove geodata")  

dt.register_event("mouse-over-image-changed", 
    toolbox_calc_distance)
